@@ -1,6 +1,7 @@
 """Interactive explorers mostly based on Bokeh."""
 import numpy as np
-from wasabi import msg as logger
+import pandas as pd
+from collections import OrderedDict
 from bokeh.plotting import figure
 from bokeh.models import CustomJS, ColumnDataSource, CDSView, IndexFilter
 from bokeh.layouts import column, row
@@ -8,22 +9,26 @@ from bokeh.palettes import Category20
 from bokeh.transform import factor_cmap
 from abc import ABC, abstractmethod
 from hover import module_config
+from hover.core import Loggable
 from hover.utils.misc import current_time
 from .local_config import bokeh_hover_tooltip
 
 
-def auto_cmap(labels):
-    """
-    Find an appropriate color map based on provide labels.
-    """
-    assert len(labels) <= 20, "Too many labels to support"
-    cmap = "Category10_10" if len(labels) <= 10 else "Category20_20"
-    return cmap
-
-
-class BokehForLabeledText(ABC):
+class BokehForLabeledText(Loggable, ABC):
     """
     Base class that keeps template explorer settings.
+
+    Assumes:
+
+    - in supplied dataframes
+      - (always) text data in a `text` column
+      - (always) xy coordinates in `x` and `y` columns
+      - (always) an index for the rows
+      - (likely) classification label in a `label` column
+
+    Does not assume:
+
+    - what the explorer serves to do.
     """
 
     DEFAULT_FIGURE_KWARGS = {
@@ -51,18 +56,19 @@ class BokehForLabeledText(ABC):
 
     DATA_KEY_TO_KWARGS = {}
 
+    MANDATORY_COLUMNS = ["text", "label", "x", "y"]
+
     def __init__(self, df_dict, **kwargs):
         """
         Operations shared by all child classes.
 
-        (1) settle the figure settings by using child class defaults + kwargs overrides
-        (2) settle the glyph settings by using child class defaults
-        (3) create widgets that child classes can override
-        (4) create data sources the correspond to class-specific data subsets.
-        (5) activate builtin search callbacks depending on the child class.
-        (6) create a (likely) blank figure under such settings
+        - settle the figure settings by using child class defaults & kwargs overrides
+        - settle the glyph settings by using child class defaults
+        - create widgets that child classes can override
+        - create data sources the correspond to class-specific data subsets.
+        - activate builtin search callbacks depending on the child class.
+        - create a (typically) blank figure under such settings
         """
-        logger.divider(f"Initializing {self.__class__.__name__}")
         self.figure_kwargs = self.__class__.DEFAULT_FIGURE_KWARGS.copy()
         self.figure_kwargs.update(kwargs)
         self.glyph_kwargs = {
@@ -76,9 +82,21 @@ class BokehForLabeledText(ABC):
         self.figure = figure(**self.figure_kwargs)
         self.reset_figure()
 
+    @classmethod
+    def from_dataset(cls, dataset, subset_mapping, *args, **kwargs):
+        """
+        Construct from a SupervisableDataset.
+        """
+        # local import to avoid import cycles
+        from hover.core.dataset import SupervisableDataset
+
+        assert isinstance(dataset, SupervisableDataset)
+        df_dict = {_v: dataset.dfs[_k] for _k, _v in subset_mapping.items()}
+        return cls(df_dict, *args, **kwargs)
+
     def reset_figure(self):
         """Start over on the figure."""
-        logger.info("Resetting figure")
+        self._info("Resetting figure")
         self.figure.renderers.clear()
 
     def _setup_widgets(self):
@@ -91,7 +109,7 @@ class BokehForLabeledText(ABC):
 
         # set up text search widgets, without assigning callbacks yet
         # to provide more flexibility with callbacks
-        logger.info("Setting up widgets")
+        self._info("Setting up widgets")
         self.search_pos = TextInput(
             title="Text contains (plain text, or /pattern/flag for regex):",
             width_policy="fit",
@@ -132,15 +150,36 @@ class BokehForLabeledText(ABC):
 
         Intended to be extended in child classes for pre/post processing.
         """
-        logger.info("Setting up dfs")
+        self._info("Setting up DataFrames")
+        supplied_keys = set(df_dict.keys())
         expected_keys = set(self.__class__.DATA_KEY_TO_KWARGS.keys())
-        assert (
-            set(df_dict.keys()) == expected_keys
-        ), f"Expected the keys of df_dict to be exactly {expected_keys}"
 
-        self.dfs = {
-            _key: (_df.copy() if copy else _df) for _key, _df in df_dict.items()
-        }
+        # perform high-level df key checks
+        supplied_not_expected = supplied_keys.difference(expected_keys)
+        expected_not_supplied = expected_keys.difference(supplied_keys)
+
+        for _key in supplied_not_expected:
+            self._warn(
+                f"{self.__class__.__name__}.__init__(): got unexpected df key {_key}"
+            )
+        for _key in expected_not_supplied:
+            self._warn(
+                f"{self.__class__.__name__}.__init__(): missing expected df key {_key}"
+            )
+
+        # create df with column checks
+        self.dfs = dict()
+        for _key, _df in df_dict.items():
+            if _key in expected_keys:
+                for _col in self.__class__.MANDATORY_COLUMNS:
+                    if not _col in _df.columns:
+                        # edge case: DataFrame has zero rows
+                        assert (
+                            _df.shape[0] == 0
+                        ), f"Missing column '{_col}' from non-empty {_key} DataFrame: found {list(_df.columns)}"
+                        _df[_col] = None
+
+                self.dfs[_key] = _df.copy() if copy else _df
 
     def _setup_sources(self):
         """
@@ -148,7 +187,7 @@ class BokehForLabeledText(ABC):
 
         Intended to be extended in child classes for pre/post processing.
         """
-        logger.info("Setting up sources")
+        self._info("Setting up sources")
         self.sources = {_key: ColumnDataSource(_df) for _key, _df in self.dfs.items()}
 
     def _update_sources(self):
@@ -158,7 +197,7 @@ class BokehForLabeledText(ABC):
         Note that it seems mandatory to re-activate the search widgets.
         This is because the source loses plotting kwargs.
         """
-        for _key in self.__class__.DATA_KEY_TO_KWARGS.keys():
+        for _key in self.dfs.keys():
             self.sources[_key].data = self.dfs[_key]
         self._activate_search_builtin()
 
@@ -169,14 +208,18 @@ class BokehForLabeledText(ABC):
 
         Note that this is a template method which heavily depends on class attributes.
         """
-        logger.info("Activating built-in search")
+        self._info("Activating built-in search")
         for _key, _dict in self.__class__.DATA_KEY_TO_KWARGS.items():
-            for _flag, _params in _dict["search"].items():
-                logger.info(
-                    f"Activated {_flag} on subset {_key} to respond to the search widgets."
-                )
-                self.glyph_kwargs[_key] = self.activate_search(
-                    self.sources[_key], self.glyph_kwargs[_key], altered_param=_params
+            if _key in self.sources.keys():
+                _responding = list(_dict["search"].keys())
+                for _flag, _params in _dict["search"].items():
+                    self.glyph_kwargs[_key] = self.activate_search(
+                        self.sources[_key],
+                        self.glyph_kwargs[_key],
+                        altered_param=_params,
+                    )
+                self._info(
+                    f"Activated {_responding} on subset {_key} to respond to the search widgets."
                 )
 
     def activate_search(self, source, kwargs, altered_param=("size", 10, 5, 7)):
@@ -299,14 +342,57 @@ class BokehForLabeledText(ABC):
         """
         pass
 
+    def auto_labels_cmap(self):
+        """
+        Find all labels and an appropriate color map.
+        """
+        labels = set()
+        for _key in self.dfs.keys():
+            labels = labels.union(set(self.dfs[_key]["label"].values))
+        labels.discard(module_config.ABSTAIN_DECODED)
+        labels = sorted(labels, reverse=True)
+
+        assert len(labels) <= 20, "Too many labels to support (max at 20)"
+        cmap = "Category10_10" if len(labels) <= 10 else "Category20_20"
+        return labels, cmap
+
+    def auto_legend_correction(self):
+        """
+        Find legend items and deduplicate by label.
+        """
+        if not hasattr(self.figure, "legend"):
+            self._fail(f"Attempting auto_legend_correction when there is no legend")
+            return
+        # extract all items and start over
+        items = self.figure.legend.items[:]
+        self.figure.legend.items.clear()
+
+        # use one item to hold all renderers matching its label
+        label_to_item = OrderedDict()
+
+        for _item in items:
+            _label = _item.label.get("value", "")
+            if not _label in label_to_item.keys():
+                label_to_item[_label] = _item
+            else:
+                label_to_item[_label].renderers.extend(_item.renderers)
+
+        # assign deduplicated items back to the legend
+        self.figure.legend.items = list(label_to_item.values())
+        return
+
 
 class BokehCorpusExplorer(BokehForLabeledText):
     """
-    Plot unlabeled, 2-D-vectorized text data points in a corpus.
+    Plot unlabeled, 2D-vectorized text data points in a corpus.
+
+    Features:
+
+    - the search widgets will highlight the results through a change of color, which is arguably the best visual effect.
     """
 
     DATA_KEY_TO_KWARGS = {
-        "raw": {
+        _key: {
             "constant": {"line_alpha": 0.4},
             "search": {
                 "size": ("size", 10, 5, 7),
@@ -314,6 +400,7 @@ class BokehCorpusExplorer(BokehForLabeledText):
                 "color": ("color", "coral", "linen", "gainsboro"),
             },
         }
+        for _key in ["raw", "train", "dev", "test"]
     }
 
     def __init__(self, df_dict, **kwargs):
@@ -325,55 +412,43 @@ class BokehCorpusExplorer(BokehForLabeledText):
         """
         super().__init__(df_dict, **kwargs)
 
-    def _setup_dfs(self, df_dict, **kwargs):
-        """Extending from the parent method."""
-        for _key in self.__class__.DATA_KEY_TO_KWARGS.keys():
-            for _col in ["text", "x", "y"]:
-                assert _col in df_dict[_key].columns
-
-        super()._setup_dfs(df_dict, **kwargs)
-
     def plot(self, *args, **kwargs):
         """
         (Re)-plot the corpus.
         Called just once per instance most of the time.
         """
-        self.figure.circle(
-            "x", "y", name="raw", source=self.sources["raw"], **self.glyph_kwargs["raw"]
-        )
+        for _key, _source in self.sources.items():
+            self.figure.circle(
+                "x", "y", name=_key, source=_source, **self.glyph_kwargs[_key]
+            )
+            self._good(f"Plotted subset {_key} with {self.dfs[_key].shape[0]} points")
 
 
 class BokehCorpusAnnotator(BokehCorpusExplorer):
     """
-    [SERVER ONLY]
     Annoate text data points via callbacks.
+
+    Features:
+
+    - alter values in the 'label' column through the widgets.
+    - **SERVER ONLY**: only works in a setting that allows Python callbacks.
     """
 
     DATA_KEY_TO_KWARGS = {
-        "raw": {
+        _key: {
             "constant": {"line_alpha": 0.3},
             "search": {
                 "size": ("size", 10, 5, 7),
-                "fill_alpha": ("fill_alpha", 0.3, 0.05, 0.2),
+                "fill_alpha": ("fill_alpha", 0.3, 0.1, 0.2),
             },
         }
+        for _key in ["raw", "train", "dev", "test"]
     }
 
     def __init__(self, df_dict, **kwargs):
         """Conceptually the same as the parent method."""
         super().__init__(df_dict, **kwargs)
 
-    def _setup_dfs(self, df_dict, **kwargs):
-        """
-        Extending from the parent method.
-
-        Add a "label" column if it is not present.
-        """
-        super()._setup_dfs(df_dict, **kwargs)
-
-        if not "label" in self.dfs["raw"].columns:
-            self.dfs["raw"]["label"] = module_config.ABSTAIN_DECODED
-
     def _layout_widgets(self):
         """Define the layout of widgets."""
         layout_rows = (
@@ -387,20 +462,21 @@ class BokehCorpusAnnotator(BokehCorpusExplorer):
         """
         Create annotator widgets and assign Python callbacks.
         """
-        from bokeh.models import TextInput, Button
+        from bokeh.models import TextInput, Button, Dropdown
 
         super()._setup_widgets()
 
-        self.annotator_input = TextInput(title="Label")
+        self.annotator_input = TextInput(title="Label:")
         self.annotator_apply = Button(
             label="Apply",
             button_type="primary",
             height_policy="fit",
             width_policy="min",
         )
-        self.annotator_export = Button(
+        self.annotator_export = Dropdown(
             label="Export",
-            button_type="success",
+            button_type="warning",
+            menu=["Excel", "CSV", "JSON", "pickle"],
             height_policy="fit",
             width_policy="min",
         )
@@ -414,35 +490,50 @@ class BokehCorpusAnnotator(BokehCorpusExplorer):
             label = self.annotator_input.value
             selected_idx = self.sources["raw"].selected.indices
             if not selected_idx:
-                logger.warn("Did not select any data points.")
+                self._warn(
+                    "Attempting annotation: did not select any data points. Eligible subset is 'raw'."
+                )
                 return
             example_old = self.dfs["raw"].at[selected_idx[0], "label"]
             self.dfs["raw"].at[selected_idx, "label"] = label
             example_new = self.dfs["raw"].at[selected_idx[0], "label"]
-            logger.good(f"Updated DataFrame, e.g. {example_old} -> {example_new}")
+            self._good(f"Updated DataFrame, e.g. {example_old} -> {example_new}")
 
             self._update_sources()
             self.plot()
-            logger.good(f"Updated annotator plot at {current_time()}")
+            self._good(f"Updated annotator plot at {current_time()}")
 
-        def callback_export(path=None):
+        def callback_export(event, path_root=None):
             """
             A callback on clicking the 'self.annotator_export' button.
 
             Saves the dataframe to a pickle.
             """
-            from dill import dump
+            export_format = event.item
 
-            # auto-determine the export path
-            if path is None:
+            # auto-determine the export path root
+            if path_root is None:
                 timestamp = current_time("%Y%m%d%H%M%S")
-                path = f".bokeh-annotated-df-{timestamp}.pkl"
+                path_root = f"hover-annotated-df-{timestamp}"
 
-            # save a pickle, then send a message
-            # note that excel/csv can be problematic with certain kinds of data
-            with open(path, "wb") as f:
-                dump(self.dfs["raw"], f)
-            logger.good(f"Saved DataFrame to {path}")
+            export_df = pd.concat(self.dfs, axis=0, sort=False, ignore_index=True)
+
+            if export_format == "Excel":
+                export_path = f"{path_root}.xlsx"
+                export_df.to_excel(export_path, index=False)
+            elif export_format == "CSV":
+                export_path = f"{path_root}.csv"
+                export_df.to_csv(export_path, index=False)
+            elif export_format == "JSON":
+                export_path = f"{path_root}.json"
+                export_df.to_json(export_path, orient="records")
+            elif export_format == "pickle":
+                export_path = f"{path_root}.pkl"
+                export_df.to_pickle(export_path)
+            else:
+                raise ValueError(f"Unexpected export format {export_format}")
+
+            self._good(f"Saved DataFrame to {export_path}")
 
         # keep the references to the callbacks
         self._callback_apply = callback_apply
@@ -459,35 +550,37 @@ class BokehCorpusAnnotator(BokehCorpusExplorer):
         Overrides the parent method.
         Determines the label->color mapping dynamically.
         """
-        all_labels = sorted(set(self.dfs["raw"]["label"].values), reverse=True)
-        cmap = auto_cmap(all_labels)
+        labels, cmap = self.auto_labels_cmap()
 
-        self.figure.circle(
-            x="x",
-            y="y",
-            name="raw",
-            color=factor_cmap("label", cmap, all_labels),
-            legend_field="label",
-            source=self.sources["raw"],
-            **self.glyph_kwargs["raw"],
-        )
+        for _key, _source in self.sources.items():
+            self.figure.circle(
+                "x",
+                "y",
+                name=_key,
+                color=factor_cmap("label", cmap, labels),
+                legend_group="label",
+                source=_source,
+                **self.glyph_kwargs[_key],
+            )
+            self._good(f"Plotted subset {_key} with {self.dfs[_key].shape[0]} points")
+
+        self.auto_legend_correction()
 
 
 class BokehSoftLabelExplorer(BokehCorpusExplorer):
     """
-    Plot text data points according to its label and confidence.
-    Currently not considering multi-label scenarios.
+    Plot text data points according to their labels and confidence scores.
+
+    Features:
+
+    - the predicted label will correspond to fill_color.
+    - the confidence score, assumed to be a float between 0.0 and 1.0, will be reflected through fill_alpha.
+    - currently not considering multi-label scenarios.
     """
 
     DATA_KEY_TO_KWARGS = {
-        "raw": {
-            "constant": {"line_alpha": 0.5},
-            "search": {"size": ("size", 10, 5, 7)},
-        },
-        "labeled": {
-            "constant": {"line_alpha": 0.5},
-            "search": {"size": ("size", 10, 5, 7)},
-        },
+        _key: {"constant": {"line_alpha": 0.5}, "search": {"size": ("size", 10, 5, 7)}}
+        for _key in ["raw", "train", "dev"]
     }
 
     def __init__(self, df_dict, label_col, score_col, **kwargs):
@@ -503,7 +596,7 @@ class BokehSoftLabelExplorer(BokehCorpusExplorer):
         kwargs.update(
             {
                 "tooltips": bokeh_hover_tooltip(
-                    label=False,
+                    label=True,
                     text=True,
                     image=False,
                     coords=True,
@@ -516,54 +609,54 @@ class BokehSoftLabelExplorer(BokehCorpusExplorer):
 
     def _setup_dfs(self, df_dict, **kwargs):
         """Extending from the parent method."""
-        for _key in self.__class__.DATA_KEY_TO_KWARGS.keys():
-            for _col in [self.label_col, self.score_col]:
-                assert _col in df_dict[_key].columns
-
         super()._setup_dfs(df_dict, **kwargs)
 
+        for _key, _df in self.dfs.items():
+            if not self.label_col in _df.columns:
+                _df[self.label_col] = module_config.ABSTAIN_DECODED
+            if not self.score_col in _df.columns:
+                _df[self.score_col] = 0.5
+
     def plot(self, **kwargs):
         """
         Plot the confidence map.
         """
+        labels, cmap = self.auto_labels_cmap()
 
-        # auto-detect all labels
-        all_labels = set()
-        for _key in self.__class__.DATA_KEY_TO_KWARGS.keys():
-            _df = self.dfs[_key]
-            _labels = set(_df[self.label_col].values)
-            all_labels = all_labels.union(_labels)
-        all_labels = sorted(all_labels, reverse=True)
-        cmap = auto_cmap(all_labels)
-
-        for _key in self.__class__.DATA_KEY_TO_KWARGS.keys():
+        for _key, _source in self.sources.items():
             # prepare plot settings
             preset_kwargs = {
-                "legend_field": self.label_col,
-                "color": factor_cmap(self.label_col, cmap, all_labels),
+                "legend_group": self.label_col,
+                "color": factor_cmap(self.label_col, cmap, labels),
                 "fill_alpha": self.score_col,
             }
             eff_kwargs = self.glyph_kwargs[_key].copy()
             eff_kwargs.update(preset_kwargs)
             eff_kwargs.update(kwargs)
 
-            self.figure.circle(
-                "x", "y", name=_key, source=self.sources[_key], **eff_kwargs
-            )
+            self.figure.circle("x", "y", name=_key, source=_source, **eff_kwargs)
+            self._good(f"Plotted subset {_key} with {self.dfs[_key].shape[0]} points")
+
+        self.auto_legend_correction()
 
 
 class BokehMarginExplorer(BokehCorpusExplorer):
     """
     Plot text data points along with two versions of labels.
     Could be useful for A/B tests.
-    Currently not considering multi-label scenarios.
+
+    Features:
+
+    - can choose to only plot the margins about specific labels.
+    - currently not considering multi-label scenarios.
     """
 
     DATA_KEY_TO_KWARGS = {
-        "raw": {
+        _key: {
             "constant": {"color": "gainsboro", "line_alpha": 0.5, "fill_alpha": 0.0},
             "search": {"size": ("size", 10, 5, 7)},
         }
+        for _key in ["raw", "train", "dev"]
     }
 
     def __init__(self, df_dict, label_col_a, label_col_b, **kwargs):
@@ -589,48 +682,52 @@ class BokehMarginExplorer(BokehCorpusExplorer):
         Plot the margins about a single label.
         """
 
-        # prepare plot settings
-        axes = ("x", "y")
-        eff_kwargs = self.glyph_kwargs["raw"].copy()
-        eff_kwargs.update(kwargs)
-        eff_kwargs["legend_label"] = f"{label}"
+        for _key, _source in self.sources.items():
+            # prepare plot settings
+            eff_kwargs = self.glyph_kwargs[_key].copy()
+            eff_kwargs.update(kwargs)
+            eff_kwargs["legend_label"] = f"{label}"
 
-        # create agreement/increment/decrement subsets
-        col_a_pos = np.where(self.dfs["raw"][self.label_col_a] == label)[0].tolist()
-        col_a_neg = np.where(self.dfs["raw"][self.label_col_a] != label)[0].tolist()
-        col_b_pos = np.where(self.dfs["raw"][self.label_col_b] == label)[0].tolist()
-        col_b_neg = np.where(self.dfs["raw"][self.label_col_b] != label)[0].tolist()
-        agreement_view = CDSView(
-            source=self.sources["raw"],
-            filters=[IndexFilter(col_a_pos), IndexFilter(col_b_pos)],
-        )
-        increment_view = CDSView(
-            source=self.sources["raw"],
-            filters=[IndexFilter(col_a_neg), IndexFilter(col_b_pos)],
-        )
-        decrement_view = CDSView(
-            source=self.sources["raw"],
-            filters=[IndexFilter(col_a_pos), IndexFilter(col_b_neg)],
-        )
-
-        to_plot = [
-            {"view": agreement_view, "marker": self.figure.square},
-            {"view": increment_view, "marker": self.figure.x},
-            {"view": decrement_view, "marker": self.figure.cross},
-        ]
-
-        # plot created subsets
-        for _dict in to_plot:
-            _view = _dict["view"]
-            _marker = _dict["marker"]
-            _marker(
-                *axes, name="raw", source=self.sources["raw"], view=_view, **eff_kwargs
+            # create agreement/increment/decrement subsets
+            col_a_pos = np.where(self.dfs[_key][self.label_col_a] == label)[0].tolist()
+            col_a_neg = np.where(self.dfs[_key][self.label_col_a] != label)[0].tolist()
+            col_b_pos = np.where(self.dfs[_key][self.label_col_b] == label)[0].tolist()
+            col_b_neg = np.where(self.dfs[_key][self.label_col_b] != label)[0].tolist()
+            agreement_view = CDSView(
+                source=_source, filters=[IndexFilter(col_a_pos), IndexFilter(col_b_pos)]
             )
+            increment_view = CDSView(
+                source=_source, filters=[IndexFilter(col_a_neg), IndexFilter(col_b_pos)]
+            )
+            decrement_view = CDSView(
+                source=_source, filters=[IndexFilter(col_a_pos), IndexFilter(col_b_neg)]
+            )
+
+            to_plot = [
+                {"view": agreement_view, "marker": self.figure.square},
+                {"view": increment_view, "marker": self.figure.x},
+                {"view": decrement_view, "marker": self.figure.cross},
+            ]
+
+            # plot created subsets
+            for _dict in to_plot:
+                _view = _dict["view"]
+                _marker = _dict["marker"]
+                _marker("x", "y", name=_key, source=_source, view=_view, **eff_kwargs)
 
 
 class BokehSnorkelExplorer(BokehCorpusExplorer):
     """
-    Plot text data points along with labeling function outputs.
+    Plot text data points along with labeling function (LF) outputs.
+
+    Features:
+
+    - each labeling function corresponds to its own line_color.
+    - uses a different marker for each type of predictions: square for 'correct', x for 'incorrect', cross for 'missed', circle for 'hit'.
+      - 'correct': the LF made a correct prediction on a point in the 'labeled' set.
+      - 'incorrect': the LF made an incorrect prediction on a point in the 'labeled' set.
+      - 'missed': the LF is capable of predicting the target class, but did not make such prediction on the particular point.
+      - 'hit': the LF made a prediction on a point in the 'raw' set.
     """
 
     DATA_KEY_TO_KWARGS = {
@@ -664,9 +761,16 @@ class BokehSnorkelExplorer(BokehCorpusExplorer):
         """Extending from the parent method."""
         super()._setup_dfs(df_dict, **kwargs)
 
-        assert "label" in self.dfs["labeled"].columns
-        if not "label" in self.dfs["raw"].columns:
-            self.dfs["raw"]["label"] = module_config.ABSTAIN_DECODED
+    def plot(self, *args, **kwargs):
+        """
+        Overriding the parent method.
+
+        Plot only the raw subset.
+        """
+        self.figure.circle(
+            "x", "y", name="raw", source=self.sources["raw"], **self.glyph_kwargs["raw"]
+        )
+        self._good(f"Plotted subset raw with {self.dfs['raw'].shape[0]} points")
 
     def plot_lf(
         self, lf, L_raw=None, L_labeled=None, include=("C", "I", "M"), **kwargs
@@ -689,7 +793,6 @@ class BokehSnorkelExplorer(BokehCorpusExplorer):
             L_labeled = self.dfs["labeled"].apply(lf, axis=1).values
 
         # prepare plot settings
-        axes = ("x", "y")
         legend_label = f"{', '.join(lf.targets)} | {lf.name}"
         color = self.palette[len(self.lfs) - 1]
 
@@ -744,10 +847,11 @@ class BokehSnorkelExplorer(BokehCorpusExplorer):
 
         # plot created subsets
         for _dict in to_plot:
+            _name = _dict["name"]
             _view = _dict["view"]
             _marker = _dict["marker"]
             _kwargs = _dict["kwargs"]
-            _marker(*axes, source=_view.source, view=_view, **_kwargs)
+            _marker("x", "y", source=_view.source, view=_view, name=_name, **_kwargs)
 
     def _view_correct(self, L_labeled):
         """
